%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.1, (the "License"); you may not use this file except in
%% compliance with the License. You should have received a copy of the
%% Erlang Public License along with this software. If not, it can be
%% retrieved via the world wide web at http://www.erlang.org/.
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Initial Developer of the Original Code is Ericsson Utvecklings AB.
%% Portions created by Ericsson are Copyright 1999, Ericsson Utvecklings
%% AB. All Rights Reserved.''
%% 
%%     $Id$
%%
%% @doc Standard functions on lists.

-module(erl.lang.list).

-export([all/2, any/2, append/2, concat/1, delete/2, drop/2,
	 dropwhile/2, duplicate/2, filter/2, find/2, first/1, flatmap/2,
	 flatten/1, flatten/2, foldl/2, foldl/3, foldr/2, foldr/3,
	 foreach/2, group/1, group/2, hd/1, intersperse/2, is_element/2,
	 is_prefix/2, is_proper/1, is_suffix/2, last/1, length/1, map/2,
	 mapfoldl/3, mapfoldr/3, max/1, merge/2, merge/3, min/1, nth/2,
	 partition/2, product/1, replace/3, reverse/1, reverse/2, seq/2,
	 seq/3, sort/1, sort/2, split/2, splitwith/2, sublist/3,
	 subtract/2, sum/1, tabulate/2, take/2, takewhile/2, tl/1,
	 transpose/1, unique/1, unzip/1, unzip3/1, zip/2, zip3/3,
	 zipwith/3, zipwith3/4]).

-import(erl.lang.list.sort).

%% Notes:
%%	* nth/2 indexes from 0, not from 1.
%%	* seq/2 and seq/3 return an empty list if the range is empty.
%%	* seq/3 can take floats as From, To and Delta.
%%
%% N.B. In general, list functions do not have a defined order of
%% evaluation, and do not necessarily visit all elements of given lists.
%% Only functions such as foreach/2, which is explicitly documented as
%% useful for producing side effects, can be relied on to visit all
%% elements, and in a particular order.

hd([H | _]) -> H;
hd(_) -> erl.lang:error(badarg).

tl([_ | T]) -> T;
tl(_) -> erl.lang:error(badarg).

length(Xs) -> length(Xs, 0).

length([_ | Xs], N) -> length(Xs, N + 1);
length([], N) -> N.

append([H | T], L) -> [H | append(T, L)];
append([], L) -> L.

subtract(L, [H | T]) -> subtract(delete(H, L), T);
subtract(L, []) -> L.

is_element(X, [X | _]) -> true;
is_element(X, [_ | Y]) -> is_element(X, Y);
is_element(_, []) -> false.

is_proper([_ | T]) -> is_proper(T);
is_proper([]) -> true;
is_proper(_) -> false.

%% Use the built-in version of reverse/2, if it exists:

reverse([] = L) -> L;
reverse([_] = L) -> L;
reverse([A | L]) -> erl.lang.list:reverse(L, [A]).

reverse([H | T], L) -> reverse(T, [H | L]);
reverse([], L) -> L.

is_prefix([X | Xs], [X | Ys]) -> is_prefix(Xs, Ys);
is_prefix([], _) -> true;
is_prefix(_, _) -> false.

is_suffix(Xs, Ys) -> is_prefix(reverse(Xs), reverse(Ys)).  % best way!

%% Flatten the (deep) list.

flatten(List) -> flatten(List, []).

%% Flatten the (deep) list and append the result onto Tail.

flatten(List, Tail) -> flatten(List, [], Tail).

flatten([H | T], C, Tail) when is_list(H) ->
    flatten(H, [T | C], Tail);
flatten([H | T], C, Tail) ->
    [H | flatten(T, C, Tail)];
flatten([], [H | C], Tail) ->
    flatten(H, C, Tail);
flatten([], [], Tail) ->
    Tail.

%% Delete the first occurance of X from the list, if one exists.

delete(X, [X | Xs]) -> Xs;
delete(X, [Y | Xs]) -> [Y | delete(X, Xs)];
delete(_, []) -> [].

%% Replace all occurences of X in the list with Y.

replace(X, Y, [X | Xs]) -> [Y | replace(X, Y, Xs)];
replace(X, Y, [Z | Xs]) -> [Z | replace(X, Y, Xs)];
replace(_, _, []) -> [].

%% Return the N:th element of (nonempty) L; N >= 0. (Origin 0.)

nth(N, L) when is_integer(N), N >= 0 -> nth_1(N, L).

nth_1(0, [H | _]) -> H;
nth_1(N, [_ | T]) -> nth_1(N - 1, T).

%% Drop N first elements from L; N >= 0 (generalisation of 'tl').

drop(N, L) when is_integer(N), N >= 1 -> drop_1(N, L);
drop(0, L) -> L.

drop_1(1, [_ | T]) -> T;
drop_1(N, [_ | T]) -> drop_1(N - 1, T).

%% Take N first elements from L; N >= 0.

take(N, L) when is_integer(N), N >= 1 -> take_1(N, L);
take(0, _) -> [].

take_1(1, [H | _]) -> [H];
take_1(N, [H | T]) -> [H | take_1(N - 1, T)].

%% Return a 2-tuple equivalent to {take(N, L), drop(N, L)}.

split(N, L) when is_integer(N), N >= 1 -> split_1(N, L, []);
split(0, L) -> {[], L}.

%% Use the built-in version of reverse/2, if it exists:

split_1(1, [H | T], As) -> {erl.lang.list:reverse(As, [H]), T};
split_1(N, [H | T], As) -> split_1(N - 1, T, [H | As]).

%% Return the last element of the (nonempty) list.

last([E]) -> E;
last([E | Es]) -> last(Es, E).

last([E | Es], _) -> last(Es, E);
last([], E) -> E.

%% Return all except the last element of the (nonempty) list.

first([_]) -> [];
first([E | Es]) -> [E | first(Es)].

%% Concatenate the lists in list L.

concat([E]) -> E;
concat([H | T]) -> H ++ concat(T);
concat([]) -> [].

%% Return the sublist from index Start to index Start + Length.

sublist(List, Start, Length) ->
    take(Length, drop(Start, List)).

%% Return 'true' if Pred(X) returns 'true' for some element X in the
%% list, and 'false' otherwise.

any(Pred, [H | T]) -> Pred(H) orelse any(Pred, T);
any(_, []) -> false. 

%% Return 'true' if Pred(X) returns 'true' for all elements X in the
%% list, and 'false' otherwise.

all(Pred, [H | T]) -> Pred(H) andalso all(Pred, T);
all(_, []) -> true. 

%% Return {value, E} for the first element E in the list such that
%% Pred(E) returns 'true', and 'none' if no such element exists.

find(Pred, [H | T]) ->
    case Pred(H) of
	true -> {value, H};
	false -> find(Pred, T)
    end;
find(_, []) -> none.

%% Return List with all elements X such that Pred(X) returns 'false'
%% removed. Pred(X) must return 'true' for all remaining elements.

filter(Pred, List) -> [E || E <- List, Pred(E)].

%% Return the list [F(E0), F(E1), ..., F(En)].

map(F, [H | T]) -> [F(H) | map(F, T)];
map(_, []) -> [].

%% Return the list concat([F(E0), F(E1), ..., F(En)]).

flatmap(F, [H | T]) -> F(H) ++ flatmap(F, T);
flatmap(_, []) -> [].

%% Return F(En, ..., F(E1, F(E0, A)) ... ).

foldl(F, A, [H | T]) -> foldl(F, F(H, A), T);
foldl(_, A, []) -> A.

%% Return F(En, ..., F(E1, E0)) ... ).

foldl(F, [H | T]) -> foldl(F, H, T).

%% Return F(E0, F(E1, ..., F(En, A) ... )).

foldr(F, A, [H | T]) -> F(H, foldr(F, A, T));
foldr(_, A, []) -> A.

%% Return F(E0, F(E1, ..., En ... )).

foldr(_, [E]) -> E;
foldr(F, [H | T]) -> F(H, foldr(F, T)).

%% Combine map & fold. F should have type (X, A) -> {X1, A1}.

mapfoldl(F, A, [H | T]) ->
    {X, A1} = F(H, A),
    {Xs, A2} = mapfoldl(F, A1, T),
    {[X | Xs], A2};
mapfoldl(_, A, []) -> {[], A}.

mapfoldr(F, A, [H | T]) ->
    {Xs, A1} = mapfoldr(F, A, T),
    {X, A2} = F(H, A1),
    {[X | Xs], A2};
mapfoldr(_, A, []) -> {[], A}.

%% Evaluate F(E0), F(E1), ..., F(En), in order.

foreach(F, [H | T]) ->
    F(H),
    foreach(F, T);
foreach(_, []) ->
    ok.

%% Return a list with N copies of X; N >= 0.

duplicate(N, X) when is_integer(N), N >= 0 -> duplicate(N, X, []).

duplicate(0, _, L) -> L;
duplicate(N, X, L) -> duplicate(N - 1, X, [X | L]).

%% Return the sequence [From, From + 1, ..., To], for integers From and
%% To. If From > To, the result is the empty list.

seq(From, To) when is_integer(From), is_integer(To), From =< To -> 
    seq_1(From, To, 1);
seq(From, To) when is_integer(From), is_integer(To) -> [].

%% Return the sequence [From, From + D, From + 2*D, ..., To], for the
%% numbers From, To and D. If D > 0 and From > To, or if D < 0 and From
%% < To, the result is the empty list. D may not be 0 unless From = To.

seq(From, To, D) when is_number(From), is_number(To), is_number(D),
		      D > 0, To >= From ->
    seq_1(From, To, D);
seq(From, To, D) when is_number(From), is_number(To), is_number(D),
		      D < 0, To =< From ->
    seq_2(From, To, D);
seq(From, From, 0) -> [From].

seq_1(From, To, D) when From =< To -> [From | seq_1(From + D, To, D)];
seq_1(_, _, _) -> [].

seq_2(From, To, D) when To =< From -> [From | seq_2(From + D, To, D)];
seq_2(_, _, _) -> [].

%% Return [F(0), F(1), ... F(N)].

tabulate(N, F) when is_integer(N), N >= 0 -> tabulate(N, F, []).

tabulate(N, F, L) when N >= 0 -> tabulate(N - 1, F, [F(N) | L]);
tabulate(_, _, L) -> L.

%% Return the sum of the elements in the list.

sum(L) -> sum(L, 0).

sum([H | T], N) -> sum(T, N + H);
sum([], N)    -> N.

%% Return the product of the elements in the list.

product(L) -> product(L, 1).

product([H | T], N) -> product(T, N * H);
product([], N)    -> N.

%% Return the minimum element of the list, in the normal term order.

min([H | T]) -> min(T, H).

min([H | T], E) when H < E -> min(T, H);
min([_ | T], E) -> min(T, E);
min([], E) -> E. 

%% Return the maximum element of the list, in the normal term order.

max([H | T]) -> max(T, H).

max([H | T], E) when H > E -> max(T, H);
max([_ | T], E) -> max(T, E);
max([], E) -> E.

%% Return list minus the longest prefix such that F(E) returns 'true'
%% for each element E.

dropwhile(Pred, [H | T] = Xs) ->
    case Pred(H) of
	true -> dropwhile(Pred, T);
	false -> Xs
    end;
dropwhile(_, []) -> [].

%% Return longest prefix of list such that F(E) returns 'true' for each
%% element E.

takewhile(Pred, [H | T]) ->
    case Pred(H) of
	true -> [H | takewhile(Pred, T)];
	false -> []
    end;
takewhile(_, []) -> [].

%% Equivalent to {takewhile(F, L), dropwhile(F, L)}.

splitwith(Pred, L) -> splitwith(Pred, L, []).

splitwith(Pred, [H | T] = Xs, As) ->
    case Pred(H) of
	true -> splitwith(Pred, T, [H | As]);
	false -> {reverse(As), Xs}
    end;
splitwith(_, [], As) ->
    {reverse(As), []}.

%% Equivalent to {filter(F, L), filter(NotF, L)}, if NotF = 'fun(X) ->
%% not F(X) end'.

partition(Pred, L) -> partition(Pred, L, [], []).

partition(Pred, [H | T], As, Bs) ->
    case Pred(H) of
	true -> partition(Pred, T, [H | As], Bs);
	false -> partition(Pred, T, As, [H | Bs])
    end;
partition(_, [], As, Bs) ->
    {reverse(As), reverse(Bs)}.

%% Return [{X0, Y0}, {X1, Y1}, ..., {Xn, Yn}] for lists [X0, X1, ...,
%% Xn] and [Y0, Y1, ..., Yn].

zip([X | Xs], [Y | Ys]) -> [{X, Y} | zip(Xs, Ys)];
zip([], []) -> [].

%% Return {[X0, X1, ..., Xn], [Y0, Y1, ..., Yn]}, for a list [{X0, Y0},
%% {X1, Y1}, ..., {Xn, Yn}].

unzip(Ts) -> unzip(Ts, [], []).

unzip([{X, Y} | Ts], Xs, Ys) -> unzip(Ts, [X | Xs], [Y | Ys]);
unzip([], Xs, Ys) -> {reverse(Xs), reverse(Ys)}.

%% Return [{X0, Y0, Z0}, {X1, Y1, Z1}, ..., {Xn, Yn, Zn}] for lists [X0,
%% X1, ..., Xn], [Y0, Y1, ..., Yn] and [Z0, Z1, ..., Zn].

zip3([X | Xs], [Y | Ys], [Z | Zs]) -> [{X, Y, Z} | zip3(Xs, Ys, Zs)];
zip3([], [], []) -> [].

%% Return {[X0, X1, ..., Xn], [Y0, Y1, ..., Yn], [Z0, Z1, ..., Zn]}, for
%% a list [{X0, Y0, Z0}, {X1, Y1, Z1}, ..., {Xn, Yn, Zn}].

unzip3(Ts) -> unzip3(Ts, [], [], []).

unzip3([{X, Y, Z} | Ts], Xs, Ys, Zs) ->
    unzip3(Ts, [X | Xs], [Y | Ys], [Z | Zs]);
unzip3([], Xs, Ys, Zs) ->
    {reverse(Xs), reverse(Ys), reverse(Zs)}.

%% Return [F(X0, Y0), F(X1, Y1), ..., F(Xn, Yn)] for lists [X0, X1, ...,
%% Xn] and [Y0, Y1, ..., Yn].

zipwith(F, [X | Xs], [Y | Ys]) -> [F(X, Y) | zipwith(F, Xs, Ys)];
zipwith(_, [], []) -> [].

%% Return [F(X0, Y0, Z0), F(X1, Y1, Z1), ..., F(Xn, Yn, Zn)] for lists
%% [X0, X1, ..., Xn], [Y0, Y1, ..., Yn] and [Z0, Z1, ..., Zn].

zipwith3(F, [X | Xs], [Y | Ys], [Z | Zs]) ->
    [F(X, Y, Z) | zipwith3(F, Xs, Ys, Zs)];
zipwith3(_, [], [], []) -> [].

%% Insert E between the elements of the list.

intersperse(_, Xs = [_]) -> Xs;
intersperse(E, [X | Xs]) -> [X, E | intersperse(E, Xs)];
intersperse(_, []) -> [].

%% Transpose rows and columns in a list of lists. Works even if sublists
%% are not of same length. Empty sublists are stripped.

transpose([]) -> [];
transpose([[] | Xss]) -> transpose(Xss);
transpose([[X | Xs] | Xss]) ->
    [[X | [H || [H | _] <- Xss]]
     | transpose([Xs | [T || [_ | T] <- Xss]])].

%% Group adjacent equals.

group(Xs) ->
    group(Xs, fun (X) -> X end).

%% Group adjacents that yield the same value for F.

group([], _F) ->
    [];
group([X | _] = Xs, F) ->
    group(Xs, F, F(X)).

group(Xs, F, P) ->
    {First, Rest} = splitwith(fun (X) -> F(X) =:= P end, Xs),
    [First | group(Rest, F)].

%% Remove adjacent duplicates from list.

unique([X | Xs]) -> [X | unique(Xs, X)];
unique([]) -> [].

unique([X | Xs], X) -> unique(Xs, X);
unique([X | Xs], _) -> [X | unique(Xs, X)];
unique([], _) -> [].

%% Merge two sorted lists.

merge([E1 | Es1], Es2 = [E2 | _]) when E1 =< E2 ->
    [E1 | merge(Es1, Es2)];
merge(Es1 = [_ | _], [E2 | Es2]) ->
    [E2 | merge(Es2, Es1)];		% switch arguments!
merge([], Es2) -> Es2;
merge(Es1, []) -> Es1.

%% Merge two sorted lists with user-defined '<' predicate.

merge(F, [E1 | Es1], Es2 = [E2 | _]) ->
    case F(E1, E2) of
	true -> [E1 | merge(F, Es1, Es2)];
	false -> [E2 | merge(F, Es2, Es1)]	% switch arguments!
    end;
merge(_, [], Es2) -> Es2;
merge(_, Es1, []) -> Es1.

%% Sort a list.

sort([X, Y | L] = L0) when X =< Y ->
    case L of
	[] -> 
	    L0;
	[Z] when Y =< Z ->
	    L0;
	[Z] when X =< Z ->
	    [X, Z, Y];
	[Z] ->
	    [Z, X, Y];
	_ ->
	    sort:split_1(X, Y, L, [], [])
    end;
sort([X, Y | L]) ->
    case L of
	[] ->
	    [Y, X];
	[Z] when X =< Z ->
	    [Y, X | L];
	[Z] when Y =< Z ->
	    [Y, Z, X];
	[Z] ->
	    [Z, Y, X];
	_ ->
	    sort:split_2(X, Y, L, [], [])
    end;
sort([_] = L) ->
    L;
sort([] = L) ->
    L.

%% Sort a list with respect to user-defined '<' predicate.

sort(Fun, L) ->
    sort:sort(Fun, L).
