%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.1, (the "License"); you may not use this file except in
%% compliance with the License. You should have received a copy of the
%% Erlang Public License along with this software. If not, it can be
%% retrieved via the world wide web at http://www.erlang.org/.
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Initial Developer of the Original Code is Ericsson Utvecklings AB.
%% Portions created by Ericsson are Copyright 1999, Ericsson Utvecklings
%% AB. All Rights Reserved.''
%% 
%%     $Id$
%%
%% @doc Standard functions for time measures. A time measure is a
%% representation of a point or a distance in time, with microsecond
%% precision. Times can be compared as normal Erlang terms, using
%% <code>==</code>, <code>/=</code>, <code>&gt;</code>,
%% <code>&gt;=</code>, etc. Arithmetic is done through the add/subtract
%% functions in this module, or by converting to seconds (micro-,
%% milli-) and back again.

-module(erl.util.time).

-export([add/2,
	 from_float/1,
	 from_micros/1,
	 from_millis/1,
	 from_seconds/1,
	 subtract/2,
	 to_float/1,
	 to_micros/1,
	 to_millis/1,
	 to_seconds/1,
	 zero/0]).

%% Time measures are always `{MegaSeconds, Seconds, MicroSeconds}'.
%% Times can be compared as normal Erlang terms, using ==, /=, >, >=,
%% etc. Arithmetic must be done using the add/subtract functions in this
%% module, or by converting to seconds (micro-, milli-) and back.

-define(MEGA, 1000000).


%% @doc Returns the zero time measure.

zero() -> {0, 0, 0}.


%% Addition/subtraction could be more efficient.
%% At least, this implementation is correct.

%% @doc Adds two times.

add(T1, T2) ->
    from_micros(to_micros(T1) + to_micros(T2)).


%% @doc Subtracts the second time from the first.

subtract(T1, T2) ->
    from_micros(to_micros(T1) - to_micros(T2)).


%% @doc Converts from time to integer microseconds.

to_micros({M,S,Mu}) ->
    (M * ?MEGA + S) * ?MEGA + Mu.


%% @doc Converts from integer microseconds to time.

from_micros(Mu) when is_integer(Mu) ->
    S = Mu div ?MEGA,
    M = S div ?MEGA,
    {M, S rem ?MEGA, Mu rem ?MEGA}.


%% @doc Converts from time to integer milliseconds.

to_millis({M,S,Mu}) ->
    (M * ?MEGA + S) * 1000 + Mu div 1000.


%% @doc Converts from integer milliseconds to time.

from_millis(Ms) when is_integer(Ms) ->
    S = Ms div 1000,
    M = S div ?MEGA,
    Mu = (Ms rem 1000) * 1000,
    {M, S rem ?MEGA, Mu}.


%% @doc Converts from time to integer seconds.

to_seconds({M,S,_Mu}) ->
    M * ?MEGA + S.


%% @doc Converts from integer seconds to time.

from_seconds(S) when is_integer(S) ->
    {S div ?MEGA, S rem ?MEGA, 0}.


%% @doc Converts from time to floating-point seconds.

to_float({M,S,Mu}) ->
    M * 1.0e6 + S + 1.0e-6 * Mu.


%% @doc Converts from floating-point seconds to time.

from_float(T) when is_float(T) ->
    M = trunc(T / 1.0e6),
    S = T - 1.0e6 * M,
    {M, trunc(S), trunc((S - trunc(S)) * ?MEGA)}.
