%% -*- erlang-indent-level: 2 -*-
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% $Id: hipe_rtl_lcm.hrl,v 1.3 2004/11/24 13:59:59 heny2875 Exp $
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%% GENERAL FUNCTIONS %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%=============================================================================
%% Returns true if the register is suitable for an identifier as
%% expression. Currently the implementation is solaris specific.
is_expr_reg([]) ->
  true;
is_expr_reg([{rtl_reg, N}|Regs]) ->
  (N >= hipe_rtl_arch:first_virtual_reg()) and is_expr_reg(Regs);
is_expr_reg([{rtl_fpreg, N}|Regs]) ->
  (N >= hipe_rtl_arch:first_virtual_reg()) and is_expr_reg(Regs);
is_expr_reg([{rtl_var, _}|_]) ->
  true. %% Henrik: Changed to true, because a var can also be an expression.

contains_machine_reg([]) ->
  false;
contains_machine_reg([{rtl_reg, N}|Regs]) ->
  (N < hipe_rtl_arch:first_virtual_reg()) orelse contains_machine_reg(Regs);
contains_machine_reg([{rtl_fpreg, N}|Regs]) ->
  (N < hipe_rtl_arch:first_virtual_reg()) orelse contains_machine_reg(Regs);
contains_machine_reg([_|Regs]) ->
  contains_machine_reg(Regs).

%%=============================================================================
%% Returns true if an RTL instruction is an expression.
%% 
is_expr(I) ->
  Defines = hipe_rtl:defines(I),
  Uses = hipe_rtl:uses(I),
  NoDefines = length(Defines),
  
  %% We don't cosider something that doesn't define anything as an expression.
  %% Also we don't consider machine registers to be expressions.
  case (NoDefines /= 0) andalso is_expr_reg(Defines) andalso
    not(contains_machine_reg(Uses)) of
    true ->
      case hipe_rtl:type(I) of
        alu ->	  
	  Dst = hipe_rtl:alu_dst(I),
	  Src1 = hipe_rtl:alu_src1(I),
	  Src2 = hipe_rtl:alu_src2(I),

	  %% Check if alu expression is untagging of boxed (rX <- vX sub 2)
	  case hipe_rtl:is_reg(Dst) andalso hipe_rtl:is_var(Src1) andalso 
	    (hipe_rtl:alu_op(I) == sub) andalso hipe_rtl:is_imm(Src2) of
	    true ->
	      case hipe_rtl:imm_value(Src2) of
		2 -> false; %% Tag for boxed. TODO: Should not be hardcoded...
		_ -> true
	      end;
	    false ->
	      true
	  end;
	       
        alub -> false; %% TODO: Split instruction to consider alu expression?
	binbase -> false;
        branch -> false;
        call -> false; %% We cannot prove that a call has no side-effects
        comment -> false;
        enter -> false;
        %fail_to -> false; %% Deprecated?
        fconv -> true;
        fixnumop -> false;
        fload -> true;
        fmove -> 
          Src = hipe_rtl:fmove_src(I),
          not (hipe_rtl:is_var(Src) orelse 
               hipe_rtl:is_reg(Src) orelse 
               hipe_rtl:is_fpreg(Src));	  
        fp -> true;
        fp_unop -> true;
        fstore -> false;
        goto -> false;
        goto_index -> false;
        gctest -> false;
        label -> false;
        load -> 
	  true;
% 	  Dst = hipe_rtl:load_dst(I),
% 	  case hipe_rtl:is_var(Dst) of
% 	    true ->
% 	      true;
% 	    false ->
% 	      false
% 	  end;
        load_address ->
	  case hipe_rtl:load_address_type(I) of
	    c_const -> false;
	    closure -> false;	%% not sure whether safe to move; 
	                        %% also probably not worth it
	    constant -> true
	  end;
        load_atom -> true;
        load_word_index -> true;
        move -> 
          Src = hipe_rtl:move_src(I),
          not (hipe_rtl:is_var(Src) orelse 
               hipe_rtl:is_reg(Src) orelse 
               hipe_rtl:is_fpreg(Src) orelse
	       hipe_rtl:is_imm(Src)); %% Doing immediates is not worth it
        multimove -> false;
        phi -> false;
        %restore_catch -> false; %% Deprecated?
        return -> false;
        store -> false;
        switch -> false
      end;
    false->
      false
  end.


%%=============================================================================
%% Replaces destination of rtl expression with empty list.
%% 
expr_set_dst(I, [Dst|Dsts]) ->
  case hipe_rtl:type(I) of
    alu -> hipe_rtl:alu_dst_update(I, Dst);
    call -> hipe_rtl:call_dstlist_update(I, [Dst|Dsts]);
    fconv -> hipe_rtl:fconv_dst_update(I, Dst);
    fload -> hipe_rtl:fload_dst_update(I, Dst);
    fmove -> hipe_rtl:fmove_dst_update(I, Dst);
    fp -> hipe_rtl:fp_dst_update(I, Dst);
    fp_unop -> hipe_rtl:fp_unop_dst_update(I, Dst);
    load -> hipe_rtl:load_dst_update(I, Dst);
    load_address -> hipe_rtl:load_address_dst_update(I, Dst);
    load_atom -> hipe_rtl:load_atom_dst_update(I, Dst);
    load_word_index -> hipe_rtl:load_word_index_dst_update(I, Dst);
    move -> hipe_rtl:move_dst_update(I, Dst);
    fixnumop -> hipe_rtl:fixnumop_dst_update(I, Dst);
    _ -> exit({?MODULE,expr_set_dst, {"bad expression"}})
  end.

%%=============================================================================
%% Replaces destination of rtl expression with empty list.
%% 
expr_clear_dst(I) ->
  case hipe_rtl:type(I) of
    alu -> hipe_rtl:alu_dst_update(I, []);
    call -> hipe_rtl:call_dstlist_update(I, []);
    fconv -> hipe_rtl:fconv_dst_update(I, []);
    fload -> hipe_rtl:fload_dst_update(I, []);
    fmove -> hipe_rtl:fmove_dst_update(I, []);
    fp -> hipe_rtl:fp_dst_update(I, []);
    fp_unop -> hipe_rtl:fp_unop_dst_update(I, []);
    load -> hipe_rtl:load_dst_update(I, []);
    load_address -> hipe_rtl:load_address_dst_update(I, []);
    load_atom -> hipe_rtl:load_atom_dst_update(I, []);
    load_word_index -> hipe_rtl:load_word_index_dst_update(I, []);
    move -> hipe_rtl:move_dst_update(I, []);
    fixnumop -> hipe_rtl:fixnumop_dst_update(I, []);
    _ -> exit({?MODULE,expr_clear_dst, {"bad expression"}})

  end.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%% PRECALC FUNCTIONS %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%=============================================================================
%% Precalculates the flow analysis and puts them in data sets for easy
%% access later.
lcm_precalc(CFG, Options) ->
  %% Calculate uses map and expression map.
  ?option_time(UsesMap = uses_map_mk(CFG), "RTL LCM uses_map_mk", Options),
  ?option_time(ExprMap = expression_map_mk(CFG), 
	       "RTL LCM calc expression_map_mk", Options),
%  Labels = hipe_rtl_cfg:reverse_postorder(CFG),
  Labels = hipe_rtl_cfg:labels(CFG),
%  StartLabel = hipe_rtl_cfg:start_label(CFG),
  AllExpr = all_expressions(CFG, Labels),
  
  %% Calculate the data sets.
  ?option_time(NodeInfo0 = mk_node_info(Labels), "RTL LCM mk_node_info", 
	       Options),
  %% ?option_time(EdgeInfo0 = mk_edge_info(), "RTL LCM mk_edge_info", 
  %%  	          Options),
  EdgeInfo0 = mk_edge_info(),
  ?option_time(NodeInfo1 = calc_up_exp(CFG, NodeInfo0, Labels), 
	       "RTL LCM calc_up_exp", Options),
  ?option_time(NodeInfo2 = calc_down_exp(CFG, NodeInfo1, Labels), 
	       "RTL LCM calc_down_exp", Options),
  ?option_time(NodeInfo3 = calc_killed_expr(CFG, NodeInfo2, UsesMap, AllExpr, 
					   Labels), 
	       "RTL LCM calc_killed_exp", Options),
  ?option_time(NodeInfo4 = calc_avail(CFG, NodeInfo3), 
	       "RTL LCM calc_avail", Options),
  ?option_time(NodeInfo5 = calc_antic(CFG, NodeInfo4, AllExpr), 
	       "RTL LCM calc_antic", Options),

  %% Earliest is calculated when needed during calc_later instead.
  %%  ?option_time(EdgeInfo1 = calc_earliest(CFG, NodeInfo5, EdgeInfo0, 
  %%					 AllExpr, Labels), 
  %%	       "RTL LCM calc_earliest", Options),
  EdgeInfo1 = EdgeInfo0,

  ?option_time({NodeInfo6, EdgeInfo2} = calc_later(CFG, NodeInfo5, EdgeInfo1,
						   AllExpr), 
	       "RTL LCM calc_later", Options),
  ?option_time(NodeInfo7 = calc_delete(CFG, NodeInfo6, Labels), 
	       "RTL LCM calc_delete", Options),
  {NodeInfo7, EdgeInfo2, AllExpr, ExprMap, Labels}.

%%%%%%%%%%%%%%%%%%%%%%% AVAIL IN/OUT FLOW ANALYSIS %%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Performs the avail in/out flow analysis.

%%=============================================================================
%% Calculates the availability.

calc_avail(CFG, NodeInfo) ->
  StartLabel = hipe_rtl_cfg:start_label(CFG),
  Work = init_work([StartLabel]),
  %% Initialize start node
  NewNodeInfo = set_avail_in(NodeInfo, StartLabel, gb_sets:empty()),
  calc_avail_fixpoint(Work, CFG, NewNodeInfo).

calc_avail_fixpoint(Work, CFG, NodeInfo) ->
  case get_work(Work) of
    fixpoint ->
      NodeInfo;
    {Label, NewWork} ->
      {NewNodeInfo, NewLabels} = calc_avail_node(Label, CFG, NodeInfo),
      NewWork2 = add_work(NewWork, NewLabels),
      calc_avail_fixpoint(NewWork2, CFG, NewNodeInfo)
  end.

calc_avail_node(Label, CFG, NodeInfo) ->
  %% Get avail in
  AvailIn = avail_in(NodeInfo, Label),

  %% Calculate avail out
  AvailOut = gb_sets:union(down_exp(NodeInfo, Label),
                           gb_sets:difference(AvailIn,
					      killed_expr(NodeInfo, Label))),

  {Changed, NodeInfo2} = 
    case avail_out(NodeInfo, Label) of
      none ->
	%% If there weren't any old avail out we use this one.
	{true, set_avail_out(NodeInfo, Label, AvailOut)};
      OldAvailOut ->
	%% Check if the avail outs are equal.
	case AvailOut == OldAvailOut of
	  true ->
	    {false, NodeInfo};
	  false ->
	    {true, set_avail_out(NodeInfo, Label, AvailOut)}
	end
    end,

  case Changed of
    true ->
      %% Update AvailIn-sets of successors and add them to worklist
      Succs = hipe_rtl_cfg:succ(hipe_rtl_cfg:succ_map(CFG), Label),
      NodeInfo3 =
	lists:foldl
	  (fun(Succ, NewNodeInfo) ->
	       case avail_in(NewNodeInfo, Succ) of
		 none ->
		   %% Initialize avail in to all expressions
		   set_avail_in(NewNodeInfo, Succ, AvailOut);

		 OldAvailIn ->
		   set_avail_in(NewNodeInfo, Succ, 
				gb_sets:intersection(OldAvailIn, AvailOut))
	       end
	   end,
	   NodeInfo2, Succs),
      {NodeInfo3, Succs};

    false ->
      {NodeInfo2, []}
  end.

%%%%%%%%%%%%%%%%%%%%%%% ANTICPABILITY IN/OUT %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% These functions handles anticipability of the basic blocks it performs
%% a precalculating fixpoint iteration and then antic_in/out is used to
%% access the precalculated data.

%%=============================================================================
%% Calculates the anticipability.

calc_antic(CFG, NodeInfo, AllExpr) ->
  % Initialize worklist with all nodes in postorder
  Labels = hipe_rtl_cfg:postorder(CFG),
  Work = init_work(Labels),

  %% Initialize start node
%  NewNodeInfo = set_antic_in(NodeInfo, StartLabel, gb_sets:empty()),
%  calc_antic_fixpoint(Work, CFG, NewNodeInfo, AllExpr).
  calc_antic_fixpoint(Work, CFG, NodeInfo, AllExpr).

calc_antic_fixpoint(Work, CFG, NodeInfo, AllExpr) ->
  case get_work(Work) of
    fixpoint ->
      NodeInfo;
    {Label, NewWork} ->
      {NewNodeInfo, NewLabels} = calc_antic_node(Label, CFG, NodeInfo, AllExpr),
      NewWork2 = add_work(NewWork, NewLabels),
      calc_antic_fixpoint(NewWork2, CFG, NewNodeInfo, AllExpr)
  end.

calc_antic_node(Label, CFG, NodeInfo, AllExpr) ->
  %% Get antic out
  AnticOut = 
    case antic_out(NodeInfo, Label) of
      none -> 
	%% DEBUG: Check if this only happens with exit nodes
	case is_exit_label(CFG, Label) of 
	  true ->
	    gb_sets:empty();
	  false ->
	    %% exit({?MODULE, calc_antic_node, 
	    %%   {"Uninitialized node not an exit label!"}}),
	    AllExpr
	end;
      
      AnticOutTemp -> AnticOutTemp
    end,

  %% Calculate antic in
  AnticIn = gb_sets:union(up_exp(NodeInfo, Label),
			  gb_sets:difference(AnticOut,
					     killed_expr(NodeInfo, Label))),

  {Changed, NodeInfo2} = 
    case antic_in(NodeInfo, Label) of
      %% If there weren't any old antic in we use this one.
      none ->
	{true, set_antic_in(NodeInfo, Label, AnticIn)};

      OldAnticIn ->
	%% Check if the antic in:s are equal.
	case AnticIn == OldAnticIn of
	  true ->
	    {false, NodeInfo};
	  false ->
	    {true, 
	     set_antic_in(NodeInfo, Label, AnticIn)}
	end
    end,

  case Changed of
    true ->
      %% Update AnticOut-sets of predecessors and add them to worklist
      Preds = hipe_rtl_cfg:pred(hipe_rtl_cfg:pred_map(CFG), Label),
      NodeInfo3 =
	lists:foldl
	  (fun(Pred, NewNodeInfo) ->
	       case antic_out(NewNodeInfo, Pred) of
		 none ->
		   %% Initialize antic out to all expressions
		   set_antic_out(NewNodeInfo, Pred, AnticIn);

		 OldAnticOut ->
		   set_antic_out(NewNodeInfo, Pred, 
				 gb_sets:intersection(OldAnticOut, AnticIn))
	       end
	   end,
	   NodeInfo2, Preds),
      {NodeInfo3, Preds};

    false ->
      {NodeInfo2, []}
  end.

%%%%%%%%%%%%%%%%% LATER / LATER IN CALCULATIONS %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Later and later in calculations. Note that Later doesn't use the ordinary
%% data structure for 

%%=============================================================================
%% Calculates the later and later in analysis.

calc_later(CFG, NodeInfo, EdgeInfo, AllExpr) ->
  StartLabel = hipe_rtl_cfg:start_label(CFG),
  Work = init_work([{node, StartLabel}]),

  %% Initialize start node
  NewNodeInfo = set_later_in(NodeInfo, StartLabel, gb_sets:empty()), 

  calc_later_fixpoint(Work, CFG, NewNodeInfo, EdgeInfo, AllExpr).

calc_later_fixpoint(Work, CFG, NodeInfo, EdgeInfo, AllExpr) ->
  case get_work(Work) of
    {{edge, From, To}, Work2} ->
      {NewNodeInfo, NewEdgeInfo, AddWork} = 
	calc_later_edge(From, To, CFG, NodeInfo, EdgeInfo, AllExpr),
      Work3 = add_work(Work2, AddWork),
      calc_later_fixpoint(Work3, CFG, NewNodeInfo, NewEdgeInfo, AllExpr);

    {{node, Label}, Work2} ->
      AddWork = calc_later_node(Label, CFG),
      Work3 = add_work(Work2, AddWork),
      calc_later_fixpoint(Work3, CFG, NodeInfo, EdgeInfo, AllExpr);

    fixpoint ->
      {NodeInfo, EdgeInfo}
  end.

calc_later_node(Label, CFG) ->
  Succs = hipe_rtl_cfg:succ(hipe_rtl_cfg:succ_map(CFG), Label),
  lists:map((fun(Succ) -> {edge, Label, Succ} end), Succs).

calc_later_edge(From, To, CFG, NodeInfo, EdgeInfo, AllExpr) ->  
  %% Instead of precalcing earliest, we calculate it when needed here
  {Earliest, EdgeInfo1} = 
    case lookup_earliest(EdgeInfo, {From, To}) of
      none ->
	IsStartLabel = (From == hipe_rtl_cfg:start_label(CFG)),
	CalcEarliest = calc_earliest_edge(NodeInfo, AllExpr, IsStartLabel,
					  From, To),
	{CalcEarliest, set_earliest(EdgeInfo, {From, To}, CalcEarliest)};

      TempEarliest ->
	{TempEarliest, EdgeInfo}
    end,

  LaterIn = later_in(NodeInfo, From),
  InvUpExp = gb_sets:difference(AllExpr, up_exp(NodeInfo, From)),
  Later = gb_sets:union(Earliest, gb_sets:intersection(LaterIn, InvUpExp)),
  
  {Changed, EdgeInfo2} =
    case lookup_later(EdgeInfo1, {From, To}) of 
      none ->
	{true, set_later(EdgeInfo1, {From, To}, Later)};

      %% FIXME use cases "Later -> " and "OldLater ->" instead of == ?
      OldLater ->
	case Later == OldLater of
	  true ->
	    {false, EdgeInfo1};
	  false ->
	    {true, set_later(EdgeInfo1, {From, To}, Later)}
	end
    end,
  
  case Changed of 
    true ->
      %% Update later in set of To-node
      case lookup_later_in(NodeInfo, To) of
	%% If the data isn't set initialize to all expressions
	none ->
 	  {set_later_in(NodeInfo, To, Later),
	   EdgeInfo2, [{node, To}]};

	OldLaterIn ->
	  NewLaterIn = gb_sets:intersection(OldLaterIn, Later),
	  %% Check if something changed
	  %% FIXME Implement faster equality test?
	  case NewLaterIn == OldLaterIn of 
	    true ->
	      {NodeInfo, EdgeInfo2, []};

	    false ->
	      {set_later_in(NodeInfo, To, NewLaterIn),
	       EdgeInfo2, [{node, To}]}
	  end
      end;

    false ->
      {NodeInfo, EdgeInfo2, []}
  end.

%%%%%%%%%%%%%%%%%% UPWARDS/DOWNWARDS EXPOSED EXPRESSIONS %%%%%%%%%%%%%%%%%%%%%%
%% Handles upwards and downwards exposed expressions.

%%=============================================================================
%% Calculates all the downwards exposed expressions in the CFG for the
%% labels given.
calc_down_exp(_CFG, NodeInfo, []) ->
  NodeInfo;
calc_down_exp(CFG, NodeInfo, [Label|Labels]) ->
  BB = hipe_rtl_cfg:bb(CFG, Label),
  Code = hipe_bb:code(BB),
  Data = gb_sets:from_list(lists:map(fun expr_clear_dst/1, exp_work(Code))),
  NewNodeInfo = set_down_exp(NodeInfo, Label, Data),
  calc_down_exp(CFG, NewNodeInfo, Labels).

%%=============================================================================
%% Calculates all the upwards exposed expressions in the CFG for the
%% labels given.
calc_up_exp(_CFG, NodeInfo, []) ->
  NodeInfo;
calc_up_exp(CFG, NodeInfo, [Label|Labels]) ->
  BB = hipe_rtl_cfg:bb(CFG, Label),
  RevCode = lists:reverse(hipe_bb:code(BB)),
  Data = gb_sets:from_list(lists:map(fun expr_clear_dst/1, exp_work(RevCode))),
  NewNodeInfo = set_up_exp(NodeInfo, Label, Data),
  calc_up_exp(CFG, NewNodeInfo, Labels).

%%=============================================================================
%% Does the work of the exposed functions.
exp_work(Code) ->
  exp_work([], Code).

exp_work([], [Instr|Instrs]) ->
  case is_expr(Instr) of
    true->
      exp_work([Instr], Instrs);
    false ->
      exp_work([], Instrs)
  end;
exp_work(Expressions, []) ->
  Expressions;
exp_work(Expressions, [Instr|Instrs]) ->
  case is_expr(Instr) of
    true ->
      NewExpressions = exp_kill_expression(Instr, [Instr|Expressions]),
      exp_work(NewExpressions, Instrs);

    false ->
      NewExpressions = exp_kill_expression(Instr, Expressions),
      exp_work(NewExpressions, Instrs)
  end.

%%=============================================================================
%% Checks if the given instruction redefines any operands of 
%% instructions in the instruction list. 
%% It returns the list of expressions with those instructions that has
%% operands redefined removed.
exp_kill_expression(_Instr, []) ->
  [];
exp_kill_expression(Instr, [CheckedExpression|Expressions]) ->
  %% Calls, gctests and stores potentially clobber everything
  case hipe_rtl:type(Instr) of
    call -> [];
    gctest -> [];
    store -> [];     %% FIXME Only regs and vars clobbered, not fregs...
    fstore -> 
      %% fstore potentially clobber float expressions
      [ExprDefine|_] = hipe_rtl:defines(CheckedExpression),
      case hipe_rtl:is_fpreg(ExprDefine) of
	true ->
	  exp_kill_expression(Instr, Expressions);
	false ->
	  [CheckedExpression | exp_kill_expression(Instr, Expressions)]
      end;
    _ ->
      InstrDefines  = hipe_rtl:defines(Instr),
      ExprUses      = hipe_rtl:uses(CheckedExpression),
      Diff          = ExprUses -- InstrDefines,

      case length(Diff) < length(ExprUses) of  
	true->
	  exp_kill_expression(Instr, Expressions);
	false ->
	  [CheckedExpression | exp_kill_expression(Instr, Expressions)]
      end
  end.
     
%%%%%%%%%%%%%%%%%%%%%%%% KILLED EXPRESSIONS %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Calculates a set of killed expressions for each block.

%%=============================================================================
%% Pre-cxalculates the killed expression set for all basic block.
calc_killed_expr(_, NodeInfo, _, _, []) ->
  NodeInfo;
calc_killed_expr(CFG, NodeInfo, UseMap, AllExpr, [Label|Labels]) ->
  BB = hipe_rtl_cfg:bb(CFG, Label),
  Code = hipe_bb:code(BB),
  KilledExprs = calc_killed_expr_bb(Code, UseMap, AllExpr, gb_sets:empty()),
  NewNodeInfo = set_killed_expr(NodeInfo, Label, KilledExprs),
  calc_killed_expr(CFG, NewNodeInfo, UseMap, AllExpr, Labels).

%%=============================================================================
%% Pre-calculates the killed expressions for a basic block.
calc_killed_expr_bb([], _UseMap, _AllExpr, KilledExprs) ->
  KilledExprs;
calc_killed_expr_bb([Instr|Instrs], UseMap, AllExpr, KilledExprs) ->
  %% Calls, gctests and stores potentially clobber everything
  case hipe_rtl:type(Instr) of
    call -> AllExpr;
    gctest -> AllExpr;
    store -> AllExpr;     %% FIXME Only regs and vars clobbered, not fregs...
    fstore -> 
      %% Kill all float expressions
      %% FIXME Make separate function is_fp_expr
      gb_sets:from_list
	(lists:foldl(fun(Expr, Fexprs) ->
			 [Define|_] = hipe_rtl:defines(Expr),
			 case hipe_rtl:is_fpreg(Define) of
			   true ->
			     [Expr|Fexprs];
			   false ->
			     Fexprs
			 end
		     end, [], gb_sets:to_list(AllExpr)));
    _ ->
      Defines = hipe_rtl:defines(Instr),
      case length(Defines) /= 0 of
	true->
	  [Define|_] = Defines,
	  NewKilledExprs = uses_map_lookup(UseMap, Define),
	  calc_killed_expr_bb(Instrs, UseMap, AllExpr,
			      gb_sets:union(NewKilledExprs, KilledExprs));
	false ->
	  calc_killed_expr_bb(Instrs, UseMap, AllExpr, KilledExprs)
      end
  end.

%%=============================================================================
%% calculates the earliest set.
% calc_earliest(_, _, EdgeInfo, _, []) ->
%   EdgeInfo;
% calc_earliest(CFG, NodeInfo, EdgeInfo, AllExpr, [From|Labels]) ->
%   IsStartLabel = (From == hipe_rtl_cfg:start_label(CFG)),
%   NewEdgeInfo = 
%     lists:foldl(fun(To, EdgeInfoAcc) ->
% 		    Earliest = 
% 		      calc_earliest_edge(NodeInfo, AllExpr, 
% 					 IsStartLabel, From, To),
% 		    set_earliest(EdgeInfoAcc, {From, To}, Earliest)
% 		end, 
% 		EdgeInfo,
% 		hipe_rtl_cfg:succ(hipe_rtl_cfg:succ_map(CFG), From)),
%   calc_earliest(CFG, NodeInfo, NewEdgeInfo, AllExpr, Labels).
  
calc_earliest_edge(NodeInfo, AllExpr, IsStartLabel, From, To) ->
  AnticIn = antic_in(NodeInfo, To),
  AvailOut = avail_out(NodeInfo, From),

  case IsStartLabel of
    true->
      gb_sets:difference(AnticIn, AvailOut);

    false ->
      AnticOut = antic_out(NodeInfo, From),
      ExprKill = killed_expr(NodeInfo, From),
      gb_sets:intersection(gb_sets:difference(AnticIn, AvailOut),
			   gb_sets:union(ExprKill, 
					 gb_sets:difference(AllExpr, 
							    AnticOut)))
  end.

%%=============================================================================
%% calculates the insert set for one edge.
calc_insert_edge(NodeInfo, EdgeInfo, From, To) ->
  Later = later(EdgeInfo, {From, To}),
  LaterIn = later_in(NodeInfo, To),	      
  gb_sets:difference(Later, LaterIn).

%%=============================================================================
%% Calculates the delete set for all basic blocks
calc_delete(_CFG, NodeInfo, []) ->
  NodeInfo;
calc_delete(CFG, NodeInfo, [Label|Labels]) ->
  case Label == hipe_rtl_cfg:start_label(CFG) of
    true -> 
      NewNodeInfo = set_delete(NodeInfo, Label, gb_sets:empty());
    false ->
      UpExp = up_exp(NodeInfo, Label),
      LaterIn = later_in(NodeInfo, Label),
      Delete = gb_sets:difference(UpExp, LaterIn),
      NewNodeInfo = set_delete(NodeInfo, Label, Delete)
  end,
  calc_delete(CFG, NewNodeInfo, Labels).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%% FIXPOINT FUNCTIONS %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%=============================================================================
%% Worklist

init_work(Labels) ->
  {Labels, [], gb_sets:from_list(Labels)}.

get_work({[Label|Left], List, Set})->
  NewWork = {Left, List, gb_sets:delete(Label, Set)},
  {Label, NewWork};
get_work({[], [], _Set}) ->
  fixpoint;
get_work({[], List, Set}) ->
  get_work({lists:reverse(List), [], Set}).

add_work(Work = {List1, List2, Set},[Label|Left])->
  case gb_sets:is_member(Label, Set) of
    true ->
      add_work(Work, Left);
    false ->
      %%io:format("Adding work: ~w\n", [Label]),
      add_work({List1, [Label|List2], gb_sets:insert(Label, Set)}, Left)
  end;
add_work(Work, []) ->
  Work.

%%=============================================================================
%% Calculates the labels that are the exit labels.
%% FIXME We do not detect dead-end loops spanning more than one block.
%%       This could potentially cause a bug in the future...
% exit_labels(CFG) ->
%   Labels = hipe_rtl_cfg:labels(CFG),
%   lists:foldl(fun(Label, ExitLabels) ->
%                   Succs = hipe_rtl_cfg:succ(hipe_rtl_cfg:succ_map(CFG), 
%                                             Label),
%                   case Succs of
% 		    [] ->
%                       [Label|ExitLabels];
% 		    [Label] -> %% Count single bb dead-end loops as exit labels
%                       [Label|ExitLabels];		      
%                     _ ->
%                       ExitLabels
%                   end
%               end, [], Labels ).

%%=============================================================================
%% Return true if label is an exit label,
%% i.e. its bb has no successors or itself as only successor.
is_exit_label(CFG, Label) ->
  case hipe_rtl_cfg:succ(hipe_rtl_cfg:succ_map(CFG), Label) of
    [] -> true;
    [Label] -> true;
    _ -> false
  end.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%% DATASET FUNCTIONS %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% The dataset is a collection of data about the CFG. 
%% It is divided into two parts, NodeInfo and EdgeInfo.
%% The pre-calculation step stores the calculated sets here.

-record(node_data, {up_exp      = none,
		    down_exp    = none,
		    killed_expr = none,
		    avail_in    = none,
		    avail_out   = none,
		    antic_in    = none,
		    antic_out   = none,
		    later_in    = none,
		    delete      = none}).

-record(edge_data, {earliest    = none,
		    later       = none,
		    insert      = none}).

%%=============================================================================
%% Creates a node info from a CFG (one entry for each Label).
mk_node_info(Labels) ->
  lists:foldl(fun(Label, DataTree) ->
		  gb_trees:insert(Label, #node_data{}, DataTree)
		  %%gb_trees:enter(Label, #node_data{}, DataTree)
	      end, 
	      gb_trees:empty(), Labels).

%%mk_edge_info(Labels) ->
%%  FIXME Should we traverse cfg and initialize edges?
mk_edge_info() ->
  gb_trees:empty().

%%=============================================================================
%% Get methods
up_exp(NodeInfo, Label) ->
  case gb_trees:lookup(Label, NodeInfo) of
    none -> 
      exit({?MODULE, up_exp, {"Node data for label not found"}}),
      none;
    {value, #node_data{up_exp = Data}} ->
      Data
  end.

down_exp(NodeInfo, Label) ->
  case gb_trees:lookup(Label, NodeInfo) of
    none -> 
      exit({?MODULE, down_exp, {"Node info not found"}}),
      none;
    {value, #node_data{down_exp = Data}} ->
      Data
  end.

killed_expr(NodeInfo, Label) ->
  case gb_trees:lookup(Label, NodeInfo) of
    none -> 
      exit({?MODULE, killed_expr, {"Node info not found"}}),
      none;
    {value, #node_data{killed_expr = Data}} ->
      Data
  end.

avail_in(NodeInfo, Label) ->
  case gb_trees:lookup(Label, NodeInfo) of
    none -> 
      exit({?MODULE, avail_in, {"Node info not found"}}),
      none;
    {value, #node_data{avail_in = Data}} ->
      Data
  end.

avail_out(NodeInfo, Label) ->
  case gb_trees:lookup(Label, NodeInfo) of
    none -> 
      exit({?MODULE, avail_out, {"Node info not found"}}),
      none;
    {value, #node_data{avail_out = Data}} ->
      Data
  end.

antic_in(NodeInfo, Label) ->
  case gb_trees:lookup(Label, NodeInfo) of
    none -> 
      exit({?MODULE, antic_in, {"Node info not found"}}),
      none;
    {value, #node_data{antic_in = Data}} ->
      Data
  end.

antic_out(NodeInfo, Label) ->
  case gb_trees:lookup(Label, NodeInfo) of
    none -> 
      exit({?MODULE, antic_out, {"Node info not found"}}),
      none;
    {value, #node_data{antic_out = Data}} ->
      Data
  end.

later_in(NodeInfo, Label) ->
  case gb_trees:lookup(Label, NodeInfo) of
    none -> 
      exit({?MODULE, later_in, {"Node info not found"}}),
      none;
    {value, #node_data{later_in = Data}} ->
      Data
  end.

lookup_later_in(NodeInfo, Label) ->
  case gb_trees:lookup(Label, NodeInfo) of
    none -> 
      none;
    {value, #node_data{later_in = Data}} ->
      Data
  end.

delete(NodeInfo, Label) ->
  case gb_trees:lookup(Label, NodeInfo) of
    none -> 
      exit({?MODULE, delete, {"Node info not found"}}),
      none;
    {value, #node_data{delete = Data}} ->
      Data
  end.

earliest(EdgeInfo, Edge) ->
  case gb_trees:lookup(Edge, EdgeInfo) of
    none -> 
      exit({?MODULE, earliest, {"Edge info not found"}}),
      none;
    {value, #edge_data{earliest = Data}} ->
      Data
  end.

lookup_earliest(EdgeInfo, Edge) ->
  case gb_trees:lookup(Edge, EdgeInfo) of
    none -> 
      none;
    {value, #edge_data{earliest = Data}} ->
      Data
  end.

later(EdgeInfo, Edge) ->
  case gb_trees:lookup(Edge, EdgeInfo) of
    none -> 
      exit({?MODULE, later, {"Edge info not found"}}),
      none;
    {value, #edge_data{later = Data}} ->
      Data
  end.

lookup_later(EdgeInfo, Edge) ->
  case gb_trees:lookup(Edge, EdgeInfo) of
    none -> 
      none;
    {value, #edge_data{later = Data}} ->
      Data
  end.

% insert(EdgeInfo, Edge) ->
%   case gb_trees:lookup(Edge, EdgeInfo) of
%     none -> 
%       exit({?MODULE, insert, {"Edge info not found"}}),
%       none;
%     {value, #edge_data{insert = Data}} ->
%       Data
%   end.

%%=============================================================================
%% Set methods
set_up_exp(NodeInfo, Label, Data) ->
  NodeData =
    case gb_trees:lookup(Label, NodeInfo) of
      none -> 
	#node_data{up_exp = Data};
      {value, OldNodeData} ->
	OldNodeData#node_data{up_exp = Data}
    end,
  gb_trees:enter(Label, NodeData, NodeInfo).

set_down_exp(NodeInfo, Label, Data) ->
  NodeData =
    case gb_trees:lookup(Label, NodeInfo) of
      none -> 
	#node_data{down_exp = Data};
      {value, OldNodeData} ->
	OldNodeData#node_data{down_exp = Data}
    end,
  gb_trees:enter(Label, NodeData, NodeInfo).

set_killed_expr(NodeInfo, Label, Data) ->
  NodeData =
    case gb_trees:lookup(Label, NodeInfo) of
      none -> 
	#node_data{killed_expr = Data};
      {value, OldNodeData} ->
	OldNodeData#node_data{killed_expr = Data}
    end,
  gb_trees:enter(Label, NodeData, NodeInfo).

set_avail_in(NodeInfo, Label, Data) ->
  NodeData =
    case gb_trees:lookup(Label, NodeInfo) of
      none -> 
	#node_data{avail_in = Data};
      {value, OldNodeData} ->
	OldNodeData#node_data{avail_in = Data}
    end,
  gb_trees:enter(Label, NodeData, NodeInfo).

set_avail_out(NodeInfo, Label, Data) ->
  NodeData =
    case gb_trees:lookup(Label, NodeInfo) of
      none -> 
	#node_data{avail_out = Data};
      {value, OldNodeData} ->
	OldNodeData#node_data{avail_out = Data}
    end,
  gb_trees:enter(Label, NodeData, NodeInfo).

set_antic_in(NodeInfo, Label, Data) ->
  NodeData =
    case gb_trees:lookup(Label, NodeInfo) of
      none -> 
	#node_data{antic_in = Data};
      {value, OldNodeData} ->
	OldNodeData#node_data{antic_in = Data}
    end,
  gb_trees:enter(Label, NodeData, NodeInfo).

set_antic_out(NodeInfo, Label, Data) ->
  NodeData =
    case gb_trees:lookup(Label, NodeInfo) of
      none -> 
	#node_data{antic_out = Data};
      {value, OldNodeData} ->
	OldNodeData#node_data{antic_out = Data}
    end,
  gb_trees:enter(Label, NodeData, NodeInfo).

set_later_in(NodeInfo, Label, Data) ->
  NodeData =
    case gb_trees:lookup(Label, NodeInfo) of
      none -> 
	#node_data{later_in = Data};
      {value, OldNodeData} ->
	OldNodeData#node_data{later_in = Data}
    end,
  gb_trees:enter(Label, NodeData, NodeInfo).

set_delete(NodeInfo, Label, Data) ->
  NodeData =
    case gb_trees:lookup(Label, NodeInfo) of
      none -> 
	#node_data{delete = Data};
      {value, OldNodeData} ->
	OldNodeData#node_data{delete = Data}
    end,
  gb_trees:enter(Label, NodeData, NodeInfo).

set_earliest(EdgeInfo, Edge, Data) ->
  EdgeData =
    case gb_trees:lookup(Edge, EdgeInfo) of
      none -> 
	#edge_data{earliest = Data};
      {value, OldEdgeData} ->
	OldEdgeData#edge_data{earliest = Data}
    end,
  gb_trees:enter(Edge, EdgeData, EdgeInfo).

set_later(EdgeInfo, Edge, Data) ->
  EdgeData =
    case gb_trees:lookup(Edge, EdgeInfo) of
      none -> 
	#edge_data{later = Data};
      {value, OldEdgeData} ->
	OldEdgeData#edge_data{later = Data}
    end,
  gb_trees:enter(Edge, EdgeData, EdgeInfo).

% set_insert(EdgeInfo, Edge, Data) ->
%   EdgeData =
%     case gb_trees:lookup(Edge, EdgeInfo) of
%       none -> 
% 	#edge_data{insert = Data};
%       {value, OldEdgeData} ->
% 	OldEdgeData#edge_data{insert = Data}
%     end,
%   gb_trees:enter(Edge, EdgeData, EdgeInfo).

%%%%%%%%%%%%%%%%%%%%%%% EXPRESSION -> USES MAP %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% This map contains a map of expression to used. Given an expression (or 
%% variable) it returns a list of expressions where that variable/expression
%% is used.

%%=============================================================================
%% creates a uses map. This map can be use to look up in which expressions
%% a variable or expression is used.
uses_map_mk(CFG) ->
  Labels = hipe_rtl_cfg:reverse_postorder(CFG),
  Map = gb_trees:empty(),
  NewMap = uses_map_mk_work(Map, CFG, Labels),
  gb_trees:balance(NewMap).

%%=============================================================================
%% Worker function for uses_map_mk goes through all blocks.
uses_map_mk_work(Map, _, []) ->
  Map;
uses_map_mk_work(Map, CFG, [Label|Labels]) ->
  BB = hipe_rtl_cfg:bb(CFG, Label),
  Code = hipe_bb:code(BB),
  NewMap = uses_map_mk_expressions(Map, Code),
  uses_map_mk_work(NewMap, CFG, Labels).

%%=============================================================================
%% Worker function for uses_map_mk goes through all expressions in a 
%% block.
uses_map_mk_expressions(Map, []) ->
  Map;
uses_map_mk_expressions(Map, [Expr|Code]) ->
  case is_expr(Expr) of
    true->
      Uses = hipe_rtl:uses(Expr),
      ExprKey = expr_clear_dst(Expr),
      NewMap = uses_map_mk_insert_uses(Map, 
                                       Uses, 
                                       ExprKey),
      uses_map_mk_expressions(NewMap, Code);
    false->
      uses_map_mk_expressions(Map, Code)
  end.

%%=============================================================================
%% Worker function for uses_map_mk inserts the uses of a expression in a
%% map.
uses_map_mk_insert_uses(Map, [], _) ->
  Map;
uses_map_mk_insert_uses(Map, [Use|Uses], Expr) ->
  case gb_trees:lookup(Use, Map) of
    {_, UseSet} ->
      NewUseSet = gb_sets:add(Expr, UseSet),
      uses_map_mk_insert_uses(gb_trees:update(Use, NewUseSet, Map),
                              Uses,
                              Expr);
    'none'->
      UseSet = gb_sets:empty(),
      NewUseSet = gb_sets:add(Expr, UseSet),
      uses_map_mk_insert_uses(gb_trees:insert(Use, NewUseSet, Map),
                              Uses,
                              Expr)
  end.

%%=============================================================================
%% Gets a gb_sets of the uses of the given Reg.
uses_map_lookup(Map, Reg) ->
  case gb_trees:lookup(Reg, Map) of
    {_, UseSet} ->
      UseSet;
    'none'->
      gb_sets:empty()
  end. 

%%%%%%%%%%%%%%%%%%%%%% EXPRESSION MAP %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% The expression map is a mapping from expression to (1) list of definitions
%% (labels where the expression is defined and the list of registers or 
%%  variables defined by an instruction in that label, represented as a 
%%  tuple {Label, Defines})) 
%% and (2) the new replacement register created for the expression

%%=============================================================================
%% Creates an expression map.
expression_map_mk(CFG) ->
  Labels = hipe_rtl_cfg:reverse_postorder(CFG),
  Map = gb_trees:empty(),
  NewMap = expression_map_mk_blocks(Map, CFG, Labels),
  gb_trees:balance(NewMap).

%%=============================================================================
%% work function for expression map.
expression_map_mk_blocks(Map, _, []) ->
  Map;
expression_map_mk_blocks(Map, CFG, [Label|Labels]) ->
  BB = hipe_rtl_cfg:bb(CFG, Label),
  Code = hipe_bb:code(BB),
  NewMap = expression_map_mk_expressions(Map, Label, Code),
  expression_map_mk_blocks(NewMap, CFG, Labels).

%%=============================================================================
%% work function for expression map.
expression_map_mk_expressions(Map, _Label, []) ->
  Map;
expression_map_mk_expressions(Map, Label, [Expr|Expressions]) ->
  case is_expr(Expr) of
    true ->
      ExprKey = expr_clear_dst(Expr),
      Defines = hipe_rtl:defines(Expr),
      case gb_trees:lookup(ExprKey, Map) of
        {value, {DefinesList, Regs}} ->
          NewMap = gb_trees:update(ExprKey, 
                                   {[{Label, Defines}|DefinesList], Regs}, 
                                   Map);
        'none' ->
          TempRegs = mk_replacement_regs(Defines),
          NewMap = gb_trees:enter(ExprKey, {[{Label, Defines}], TempRegs}, Map)
      end,
      expression_map_mk_expressions(NewMap, Label, Expressions);
    false ->
      expression_map_mk_expressions(Map, Label, Expressions)
  end.

%%=============================================================================
%% Creates new temporaries to replace defines in moved expressions.
mk_replacement_regs([]) ->
  [];
mk_replacement_regs(Defines) ->
  mk_replacement_regs(Defines, []).

mk_replacement_regs([], NewRegs) ->
  lists:reverse(NewRegs);
mk_replacement_regs([Define|Defines], NewRegs) ->
  case hipe_rtl:is_reg(Define) of
    true ->
      mk_replacement_regs(Defines, [hipe_rtl:mk_new_reg()|NewRegs]);
    false ->
      case hipe_rtl:is_var(Define) of
	true ->
	  mk_replacement_regs(Defines, [hipe_rtl:mk_new_var()|NewRegs]);
	false ->
	  case hipe_rtl:is_fpreg(Define) of
	    true ->
	      mk_replacement_regs(Defines, [hipe_rtl:mk_new_fpreg()|NewRegs])
	  end
      end
  end.
  
%%=============================================================================
%% Performs a lookup and returns the expression of the label Reg.
expression_map_lookup(Map, ExprKey) ->
  gb_trees:lookup(ExprKey, Map).

%%=============================================================================
%% Gets the actual code to be generated for insertions of an expression.
expression_map_get_instr(Map, ExprKey) ->
  case gb_trees:lookup(ExprKey, Map) of
    {value, {_, Regs}} ->
      expr_set_dst(ExprKey, Regs);
    none ->
      exit({?MODULE,expression_map_get_instr, {"expression missing"}})
  end.

%%=============================================================================
%% Create code that moves a value
mk_expression_move_instr([Reg], [Define]) ->
  case hipe_rtl:is_fpreg(Reg) of
    true ->
      hipe_rtl:mk_fmove(Reg, Define);
    false ->
      %% FIXME Check is_var() orelse is_reg() ?
      hipe_rtl:mk_move(Reg, Define)
  end;
mk_expression_move_instr([Reg|Regs], Defines) ->
  %% FIXME Does this really work? What about floats...
  hipe_rtl:mk_multimove([Reg|Regs], Defines);
mk_expression_move_instr(_, []) ->
  exit({?MODULE,mk_expression_move_instr, {"bad match"}}).

%%=============================================================================
%% Returns a set of all expressions in the code.
all_expressions(_CFG, []) ->
  gb_sets:empty();
all_expressions(CFG, [Label|Labels]) ->
  BB = hipe_rtl_cfg:bb(CFG, Label),
  Code = hipe_bb:code(BB),
  gb_sets:union(all_expressions_bb(Code),
                all_expressions(CFG, Labels)).

%%=============================================================================
%% Returns a set of expressions in a basic block.
all_expressions_bb([]) ->
  gb_sets:empty();
all_expressions_bb([Expr|Exprs]) ->
  case is_expr(Expr) of
    true->
      ExprKey = expr_clear_dst(Expr),
      ExprSet = all_expressions_bb(Exprs),
      gb_sets:add(ExprKey, ExprSet);
    false ->
      all_expressions_bb(Exprs)
  end.

%%%%%%%%%%%%%%%%%% NEW BB (INSERT BETWEEN) MAP %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Map from edges to labels
%%

%%=============================================================================
%% 
mk_new_bb_map() ->
  gb_trees:empty().

new_bb_map_insert(Map, Edge, Label) ->
  gb_trees:enter(Edge, Label, Map).

new_bb_map_lookup(Map, Edge) ->
  gb_trees:lookup(Edge, Map).
