%%% $Id: hipe_vector.erl,v 1.2 2004/01/19 11:21:50 richardc Exp $
%%% A thin abstraction layer which permits plugging in alternative
%%% implementations in place of BEAM's built-in 'vector:' operations.

-module(hipe_vector).

-export([new/2, from_list/1, to_list/1, get/2, set/3]).

%%% Uncomment exactly one of these
-define(VECTOR_USING_BEAM,1).
%-define(VECTOR_USING_GBTREES,1).
%-define(VECTOR_USING_TUPLES,1).

-ifdef(VECTOR_USING_BEAM).
%%%
%%% Using BEAM's built-in 'vector:' operations.
%%%
new(Size, Init) ->
    vector:new(Size, Init).
from_list(List) ->
    vector:from_list(List).
to_list(Vec) ->
    vector:to_list(Vec).
get(Ix, Vec) ->
    vector:get(Ix, Vec).
set(Ix, Vec, Val) ->
    vector:set(Ix, Vec, Val).
-endif.

-ifdef(VECTOR_USING_GBTREES).
%%%
%%% BEAM-like 'vector:' implementation on top of gb_trees.
%%% All valid keys are always present in the tree, since this simplifies
%%% the to_list/1, get/2, and set/3 operations.
%%%
new(Size, Init) ->
    hipe_vector:from_list_gb(lists:duplicate(Size, Init)).
from_list(List) ->
    hipe_vector:from_list_gb(List).
to_list(Vec) ->
    gb_trees:values(Vec).
get(Ix, Vec) ->
    gb_trees:get(Ix, Vec).
set(Ix, Vec, Val) ->
    gb_trees:update(Ix, Val, Vec).
from_list_gb(List) ->
    from_list_gb(List, 1, []).
from_list_gb([Value|Values], Index, RevOrdDict) ->
    from_list_gb(Values, Index+1, [{Index,Value}|RevOrdDict]);
from_list_gb([], _, RevOrdDict) ->
    gb_trees:from_orddict(lists:reverse(RevOrdDict)).
-endif.


%%%
%%% BEAM-like 'vector:' implementation on top of ordinary tuples.
%%% This is a trivial reference implementation with a very slow set/3
%%% operation.
%%% Disabled by default.
%%%
-ifdef(VECTOR_USING_TUPLES).
new(Size, Init) ->
    list_to_tuple(lists:duplicate(Size, Init)).
from_list(List) ->
    list_to_tuple(List).
to_list(Vec) ->
    tuple_to_list(Vec).
get(Ix, Vec) ->
    element(Ix, Vec)).
set(Ix, Vec, Val) ->
    setelement(Ix, Vec, Val)).
-endif.
